.navbar {
  display: none;
  border: none;
}

.infobar {
  display: none;
  border: none;
}

.size_box {
  display: none;
  border: none;
}

.visualisation {
  position:absolute;
  top: 60px;
  border: 1px solid #aaa;
  border-radius: 10px;
  padding: 5px;
  width: 90%;
  height: 600px;
  background-color: white;
  border: none;
}

.chart_rect {
  fill: white;
}
